#include "..\\uLilith\\Plugin\\GenericPlugin.h"
#include <shlwapi.h>
#include "Resource.h"
#include <process.h>

#include <string>
#include <xstring>
#include <map>


#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "gdiplus.lib")
#pragma comment(lib, "winmm.lib")

#ifdef __INTEL_COMPILER
#pragma warning(disable:1028)
#pragma warning(disable:1738)
#pragma warning(disable:1744)
#endif

#define STR_LEN 1024
#define PROP_EFFECTDATA TEXT("Prop::EffectData")

#define LARGE_BUFFER_LEN 65536

namespace std
{
	class string_less_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class wstring_less_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class string_greater_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) > 0);
		}
	};
	class wstring_greater_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) > 0);
		}
	};
};
#ifdef _UNICODE
#define _tstring std::wstring
#define _tstring_less_i std::wstring_less_i
#else
#define _tstring std::string
#define _tstring_less_i std::string_less_i
#endif


inline BOOL NVStringToBool(LPCTSTR str)
{
	if (_tcsicmp(str, TEXT("TRUE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("ENABLE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("On")) == 0)
	{
		return TRUE;
	}
	else if (_tstoi(str) == 1)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
inline BOOL GetPrivateProfileBool(LPCTSTR pszField, LPCTSTR pszKey, BOOL fDefault, LPCTSTR pszFileName)
{
	TCHAR szBuf[256];

	GetPrivateProfileString(pszField, pszKey, fDefault ? TEXT("TRUE") : TEXT("FALSE"), szBuf, 256, pszFileName);
	return NVStringToBool(szBuf);
}

enum EffectType
{
	EffectClear = 0,
	EffectMosaic,
	EffectMax,
};

struct EffectEntry
{
	EffectType type;
	int nMosaicSize;
};
typedef std::map<_tstring, EffectEntry, _tstring_less_i> EffectList;

typedef std::map<_tstring, EffectList> FaceEffectList;
struct EffectData
{
	FaceEffectList faceList;
	HWND hwndProperty;
};


extern "C"
{
	void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info);
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks);
void WINAPI UninitializePlugin(LPVOID lpvUserData);
HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner);
int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
void WINAPI FaceEffectCallback(LPVOID lpvUserData, HANDLE hFace, const GenericPlugin::DrawItem& Canvas, LPCTSTR pszFieldName);

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam);

static void ProcessClear(const GenericPlugin::DrawItem& Canvas);
static void ProcessMosaic(const GenericPlugin::DrawItem& Canvas, int nMosaicSize);

GenericPlugin::GenericPluginInfo g_Info =
{
	0x200,
	TEXT("SampleFaceEffect"),
	TEXT("Northern Verse (Project9k)"),
	TEXT("tFCXGtFNgvOC̃Tvł"),
	InitializePlugin,
	UninitializePlugin,
	ShowProperty,
	OnFaceLoad,
	OnFaceUnload,
	NULL,//OnSoundOpen,
	NULL,//OnSoundChange,
	NULL,//OnSoundClose,
	NULL,//OnUpdateVisual,
	NULL,//OnSoundStatus,
	FaceEffectCallback,
};

GenericPlugin::PluginCallbacks *g_callbacks = NULL;
HINSTANCE g_hThisInstance = NULL;

BOOL APIENTRY DllMain(HINSTANCE hThisInstance, DWORD dwNotification, LPVOID lpReserved)
{
	if (dwNotification == DLL_PROCESS_ATTACH)
	{
		g_hThisInstance = hThisInstance;
		DisableThreadLibraryCalls(hThisInstance);
	}
	else if (dwNotification == DLL_PROCESS_DETACH)
	{
	}

	return TRUE;
}



void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info)
{
	info = g_Info;
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks)
{
	EffectData *data = new EffectData;
	data->faceList.clear();
	data->hwndProperty = NULL;

	g_callbacks = (GenericPlugin::PluginCallbacks *) callbacks;

	return data;
}

void WINAPI UninitializePlugin(LPVOID lpvUserData)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		data->faceList.clear();
		delete data;
	}
}

HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		if (data->hwndProperty)
		{
			SetForegroundWindow(data->hwndProperty);
			return data->hwndProperty;
		}
		else
		{
			data->hwndProperty = CreateDialogParam(g_hThisInstance, MAKEINTRESOURCE(IDD_PROPERTY), hwndOwner, PropertyDialogProc, (LPARAM) data);
			return data->hwndProperty;
		}
	}
	return NULL;
}

int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		FaceEffectList::iterator it;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		it = data->faceList.find(pszFaceName);
		if (it == data->faceList.end())
		{
			data->faceList.insert(make_pair(pszFaceName, EffectList()));
			it = data->faceList.find(pszFaceName);
		}
		if (it != data->faceList.end())
		{
			TCHAR szFieldName[STR_LEN];
			HANDLE hffi = g_callbacks->FindFirstFaceItem(hFace, TEXT("SampleMosaic"), szFieldName, STR_LEN);
			if (hffi)
			{
				do
				{
					EffectEntry mosaic;
					mosaic.type = EffectMosaic;
					mosaic.nMosaicSize = GetPrivateProfileInt(szFieldName, TEXT("MosaicSize"), 0, pszFileName);
					it->second.insert(std::make_pair(szFieldName, mosaic));
				}
				while (g_callbacks->FindNextFaceItem(hffi, szFieldName, STR_LEN));
				g_callbacks->FindFaceItemClose(hffi);
			}
			hffi = g_callbacks->FindFirstFaceItem(hFace, TEXT("SampleClear"), szFieldName, STR_LEN);
			if (hffi)
			{
				do
				{
					EffectEntry clear;
					clear.type = EffectClear;
					it->second.insert(std::make_pair(szFieldName, clear));
				}
				while (g_callbacks->FindNextFaceItem(hffi, szFieldName, STR_LEN));
				g_callbacks->FindFaceItemClose(hffi);
			}
		}
	}

	return 0;
}

int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		FaceEffectList::iterator it;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		it = data->faceList.find(pszFaceName);
		if (it != data->faceList.end())
		{
			data->faceList.erase(it);
		}
	}

	return 0;
}

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	switch (uMessage)
	{
	case WM_INITDIALOG:
		{
			EffectData *data = (EffectData *) lParam;
			SetProp(hwnd, PROP_EFFECTDATA, (HANDLE) data);
			g_callbacks->RegisterDialog(hwnd);
			ShowWindow(hwnd, SW_SHOWNORMAL);
		}
		break;
	case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
			case IDOK:
			case IDCANCEL:
				{
					DestroyWindow(hwnd);
				}
				break;
			}
		}
		break;
	case WM_DESTROY:
		{
			EffectData *data = (EffectData *) GetProp(hwnd, PROP_EFFECTDATA);
			g_callbacks->UnregisterDialog(hwnd);
			if (data)
			{
				data->hwndProperty = NULL;
			}
		}
		break;
	}

	return FALSE;
}


void WINAPI FaceEffectCallback(LPVOID lpvUserData, HANDLE hFace, const GenericPlugin::DrawItem& Canvas, LPCTSTR pszFieldName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		EffectList::iterator it;
		FaceEffectList::iterator itList;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		itList = data->faceList.find(pszFaceName);
		if (itList != data->faceList.end())
		{
			it = itList->second.find(pszFieldName);
			if (it != itList->second.end())
			{
				if (it->second.type == EffectClear)
				{
					ProcessClear(Canvas);
				}
				else if (it->second.type == EffectMosaic)
				{
					int nMosaicSize = it->second.nMosaicSize;
					if (nMosaicSize > 0)
					{
						ProcessMosaic(Canvas, nMosaicSize);
					}
				}
			}
		}
	}
}

static void ProcessClear(const GenericPlugin::DrawItem& Canvas)
{
	Gdiplus::BitmapData bits;
	Gdiplus::Rect rc(Canvas.rcView.left, Canvas.rcView.top, Canvas.rcView.right - Canvas.rcView.left, Canvas.rcView.bottom - Canvas.rcView.top);

	Gdiplus::Status ret = Canvas.FrameBuffer->LockBits(&rc, Gdiplus::ImageLockModeRead | Gdiplus::ImageLockModeWrite, PixelFormat32bppARGB, &bits);
	if (ret == Gdiplus::Ok)
	{
		if (bits.PixelFormat == PixelFormat32bppARGB)
		{
			UINT i;

			for (i = 0; i < bits.Height; i++)
			{
				ZeroMemory(((LPBYTE) bits.Scan0) + i * bits.Stride, bits.Width * (32 / 8));
			}
		}

		Canvas.FrameBuffer->UnlockBits(&bits);
	}
}

static void ProcessMosaic(const GenericPlugin::DrawItem& Canvas, int nMosaicSize)
{
	Gdiplus::BitmapData bits;
	Gdiplus::Rect rc(Canvas.rcView.left, Canvas.rcView.top, Canvas.rcView.right - Canvas.rcView.left, Canvas.rcView.bottom - Canvas.rcView.top);

	Gdiplus::Status ret = Canvas.FrameBuffer->LockBits(&rc, Gdiplus::ImageLockModeRead | Gdiplus::ImageLockModeWrite, PixelFormat32bppARGB, &bits);
	if (ret == Gdiplus::Ok)
	{
		if (bits.PixelFormat == PixelFormat32bppARGB)
		{
			UINT i, j;
			int m, n;
			UINT iMax = bits.Height / nMosaicSize;
			UINT jMax = bits.Width / nMosaicSize;

			for (i = 0; i < iMax; i++)
			{
				for (j = 0; j < jMax; j++)
				{
					UINT r = 0, g = 0, b = 0;

					for (m = 0; m < nMosaicSize; m++)
					{
						LPBYTE pCur = ((LPBYTE) bits.Scan0) + (i * nMosaicSize + m) * bits.Stride + j * nMosaicSize * (32 / 8);

						for (n = 0; n < nMosaicSize; n++)
						{
							r += pCur[n * (32 / 8) + 0];
							g += pCur[n * (32 / 8) + 1];
							b += pCur[n * (32 / 8) + 2];
						}
					}
					{
						int div = nMosaicSize * nMosaicSize;

						r /= div;
						g /= div;
						b /= div;
					}

					for (m = 0; m < nMosaicSize; m++)
					{
						LPBYTE pCur = ((LPBYTE) bits.Scan0) + (i * nMosaicSize + m) * bits.Stride + j * nMosaicSize * (32 / 8);

						for (n = 0; n < nMosaicSize; n++)
						{
							pCur[n * (32 / 8) + 0] = r;
							pCur[n * (32 / 8) + 1] = g;
							pCur[n * (32 / 8) + 2] = b;
						}
					}
				}
			}
		}
		Canvas.FrameBuffer->UnlockBits(&bits);
	}
}
